@@ -41,6 +41,7 @@ from ..exception import (
     FieldError,
     UniqueError,
     CloseError,
+    PlaceholderError,
 )
 
 from ..compat import *
@@ -64,31 +65,6 @@ class SQLiteConnection(SQLConnection, sqlite3.Connection):
         self.closed = 1
         return r
 
-#         self.cursors = set()
-# 
-#     def cursor(self, *args, **kwargs):
-#         """https://docs.python.org/3.11/library/sqlite3.html#cursor-objects"""
-#         cur = super().cursor(*args, **kwargs)
-#         self.cursors.add(cur)
-#         return cur
-# 
-#     def close(self, *args, **kwargs):
-#         for cur in self.cursors:
-#             try:
-#                 cur.close()
-# 
-#             except Exception:
-#                 pass
-# 
-#         r = super().close(*args, **kwargs)
-#         self.closed = 1
-#         return r
-# 
-#     def _execute(self, query_str):
-#         pout.v(self.in_transaction)
-#         return super()._execute(query_str)
-#         pout.v(self.in_transaction)
-
 
 class BooleanType(object):
     FIELD_TYPE = 'BOOL'
@@ -197,17 +173,19 @@ class DictType(object):
 
 
 class SQLite(SQLInterface):
-
-    val_placeholder = '?'
+    """
+    https://docs.python.org/3/library/sqlite3.html
+    """
+    LIMIT_NONE = -1
 
     _connection = None
 
     @classmethod
-    def configure(cls, connection_config):
-        dsn = getattr(connection_config, 'dsn', '')
+    def configure(cls, config):
+        dsn = getattr(config, 'dsn', '')
         if dsn:
-            host = connection_config.host
-            db = connection_config.database
+            host = config.host
+            db = config.database
             if not host:
                 path = db
 
@@ -218,19 +196,25 @@ class SQLite(SQLInterface):
                 path = os.sep.join([host, db])
 
         else:
-            path = connection_config.database
+            path = config.database
 
         if not path:
-            raise ValueError("no sqlite db path found in connection_config")
+            raise ValueError("no sqlite db path found in config")
 
-        connection_config.path = path
-        return connection_config
+        config.path = path
+        return config
 
-    def _connect(self, connection_config):
+    def get_paramstyle(self):
+        """
+        https://docs.python.org/3/library/sqlite3.html#sqlite3.paramstyle
+        """
+        return sqlite3.paramstyle
+
+    def _connect(self, config):
         """
         https://docs.python.org/3.11/library/sqlite3.html#sqlite3.connect
         """
-        path = connection_config.path
+        path = config.path
 
         # https://docs.python.org/2/library/sqlite3.html#default-adapters-and-converters
         options = {
@@ -244,8 +228,8 @@ class SQLite(SQLInterface):
         }
         option_keys = list(options.keys()) + ['timeout', 'cached_statements']
         for k in option_keys:
-            if k in connection_config.options:
-                options[k] = connection_config.options[k]
+            if k in config.options:
+                options[k] = config.options[k]
 
         try:
             self._connection = sqlite3.connect(path, **options)
@@ -291,12 +275,12 @@ class SQLite(SQLInterface):
     def _configure_connection(self, **kwargs):
         # turn on foreign keys
         # http://www.sqlite.org/foreignkeys.html
-        self._query('PRAGMA foreign_keys = ON', ignore_result=True, **kwargs)
+        self._raw('PRAGMA foreign_keys = ON', ignore_result=True, **kwargs)
 
         # by default we can read/write, so only bother to run this query if we need to
         # make the connection actually readonly
-        if self.connection_config.readonly:
-            self._readonly(self.connection_config.readonly, **kwargs)
+        if self.config.readonly:
+            self._readonly(self.config.readonly, **kwargs)
 
     def _get_connection(self):
         return self._connection
@@ -306,7 +290,7 @@ class SQLite(SQLInterface):
         self._connection = None
 
     def _readonly(self, readonly, **kwargs):
-        self._query(
+        self._raw(
             # https://stackoverflow.com/a/49630725/5006
             'PRAGMA query_only = {}'.format("ON" if readonly else "OFF"),
             ignore_result=True,
@@ -321,7 +305,7 @@ class SQLite(SQLInterface):
             query_str += ' AND name = ?'
             query_args.append(str(table_name))
 
-        ret = self._query(query_str, *query_args, **kwargs)
+        ret = self._raw(query_str, *query_args, **kwargs)
         return [r['tbl_name'] for r in ret]
 
     def _get_indexes(self, schema, **kwargs):
@@ -330,21 +314,21 @@ class SQLite(SQLInterface):
         # http://www.mail-archive.com/sqlite-users@sqlite.org/msg22055.html
         # http://stackoverflow.com/questions/604939/
         ret = {}
-        rs = self._query('PRAGMA index_list({})'.format(self._normalize_table_name(schema)), **kwargs)
+        rs = self._raw('PRAGMA index_list({})'.format(self.render_table_name_sql(schema)), **kwargs)
         if rs:
             for r in rs:
                 iname = r['name']
                 ret.setdefault(iname, [])
-                indexes = self._query('PRAGMA index_info({})'.format(r['name']), **kwargs)
+                indexes = self._raw('PRAGMA index_info({})'.format(r['name']), **kwargs)
                 for idict in indexes:
                     ret[iname].append(idict['name'])
 
         return ret
 
     def _delete_tables(self, **kwargs):
-        self._query('PRAGMA foreign_keys = OFF', ignore_result=True, **kwargs);
+        self._raw('PRAGMA foreign_keys = OFF', ignore_result=True, **kwargs);
         ret = super()._delete_tables(**kwargs)
-        self._query('PRAGMA foreign_keys = ON', ignore_result=True, **kwargs);
+        self._raw('PRAGMA foreign_keys = ON', ignore_result=True, **kwargs);
         return ret
 
     def _delete_table(self, schema, **kwargs):
@@ -352,10 +336,11 @@ class SQLite(SQLInterface):
         https://www.sqlite.org/lang_droptable.html
         """
         #query_str = 'DROP TABLE IF EXISTS {}'.format(str(schema))
-        query_str = "DROP TABLE IF EXISTS {}".format(self._normalize_table_name(schema))
-        ret = self.query(query_str, ignore_result=True, **kwargs)
+        query_str = "DROP TABLE IF EXISTS {}".format(self.render_table_name_sql(schema))
+        self._raw(query_str, ignore_result=True, **kwargs)
 
     def create_error(self, e, **kwargs):
+        kwargs.setdefault("error_module", sqlite3)
         if isinstance(e, sqlite3.OperationalError):
             e_msg = str(e)
             if "no such column" in e_msg or "has no column" in e_msg:
@@ -383,6 +368,28 @@ class SQLite(SQLInterface):
             if "closed database" in e_msg:
                 e = CloseError(e)
 
+            elif "Incorrect number of bindings supplied" in e_msg:
+                e = PlaceholderError(e, *kwargs.get("error_args", []))
+
+            else:
+                e = super().create_error(e, **kwargs)
+
+        elif isinstance(e, sqlite3.InterfaceError):
+            e_msg = str(e)
+            if "Error binding parameter" in e_msg and "error_args" in kwargs:
+                error_args = kwargs["error_args"]
+                ms = re.search(r"parameter\s(\d+)", e_msg)
+                index = int(ms.group(1))
+                value = error_args[1][index]
+                e = PlaceholderError(
+                    e,
+                    *error_args,
+                    message=f"Placeholder {index} of query has unexpected type {type(value)}",
+                )
+
+            else:
+                e = super().create_error(e, **kwargs)
+
         else:
             e = super().create_error(e, **kwargs)
 
@@ -391,12 +398,12 @@ class SQLite(SQLInterface):
     def _get_fields(self, table_name, **kwargs):
         """return all the fields for the given table"""
         ret = {}
-        query_str = 'PRAGMA table_info({})'.format(self._normalize_table_name(table_name))
-        fields = self._query(query_str, **kwargs)
+        query_str = 'PRAGMA table_info({})'.format(self.render_table_name_sql(table_name))
+        fields = self._raw(query_str, **kwargs)
         #pout.v([dict(d) for d in fields])
 
-        query_str = 'PRAGMA foreign_key_list({})'.format(self._normalize_table_name(table_name))
-        fks = {f["from"]: f for f in self._query(query_str, **kwargs)}
+        query_str = 'PRAGMA foreign_key_list({})'.format(self.render_table_name_sql(table_name))
+        fks = {f["from"]: f for f in self._raw(query_str, **kwargs)}
         #pout.v([dict(d) for d in fks.values()])
 
         pg_types = {
@@ -442,7 +449,7 @@ class SQLite(SQLInterface):
 
         return ret
 
-    def _normalize_date_SQL(self, field_name, field_kwargs, symbol):
+    def render_date_field_sql(self, field_name, field_kwargs, symbol):
         """
         allow extracting information from date
 
@@ -462,11 +469,11 @@ class SQLite(SQLInterface):
         }
 
         for k, v in field_kwargs.items():
-            fstrs.append([k_opts[k].format(self._normalize_name(field_name)), self.val_placeholder, v])
+            fstrs.append([k_opts[k].format(self.render_field_name_sql(field_name)), self.PLACEHOLDER, v])
 
         return fstrs
 
-    def _normalize_sort_SQL(self, field_name, field_vals, sort_dir_str):
+    def render_sort_field_sql(self, field_name, field_vals, sort_dir_str):
         """
         allow sorting by a set of values
 
@@ -479,31 +486,16 @@ class SQLite(SQLInterface):
         else:
             fvi = (t for t in enumerate(reversed(field_vals))) 
 
-        query_sort_str = ['  CASE {}'.format(self._normalize_name(field_name))]
+        query_sort_str = ['  CASE {}'.format(self.render_field_name_sql(field_name))]
         query_args = []
         for i, v in fvi:
-            query_sort_str.append('    WHEN {} THEN {}'.format(self.val_placeholder, i))
+            query_sort_str.append('    WHEN {} THEN {}'.format(self.PLACEHOLDER, i))
             query_args.append(v)
 
         query_sort_str.append('  END')
         query_sort_str = "\n".join(query_sort_str)
         return query_sort_str, query_args
 
-    def _normalize_bounds_SQL(self, bounds, sql_options):
-        offset = bounds.offset
-        if sql_options.get('one_query', False):
-            limit = 1
-
-        else:
-            limit, offset = bounds.get()
-            if not bounds.has_limit():
-                limit = -1
-
-        return 'LIMIT {} OFFSET {}'.format(
-            limit,
-            offset
-        )
-
     def render_datatype_int_sql(self, field_name, field, **kwargs):
         if field.is_pk():
             field_type = 'INTEGER PRIMARY KEY'
